/*
 * Copyright 2010 Martin Schreiber
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef CLBMVISUALIZATION_HPP
#define CLBMVISUALIZATION_HPP

#include "CLbmOpenCl.hpp"
#include "libmath/CVector.hpp"
#include "CGui.hpp"

template <typename T>
class CLbmVisualization
{
	T *velocity;
	T *density;
	int *flags;
	CVector<3,int> domain_cells;

	CLbmOpenCl<T> *cLbmOpencl;

	// rendering parameters
	CVector<3,float> domain_scale;
//	CVector<3,float> translate_domain_cells;

public:
	CLbmVisualization()
		: velocity(NULL),
	density(NULL),
	flags(NULL)
	{
	}

	void setup(CLbmOpenCl<T> &p_cLbmOpencl)
	{
		cLbmOpencl = &p_cLbmOpencl;

		domain_cells = p_cLbmOpencl.domain_cells;

		delete [] velocity;
		velocity = new T[domain_cells.elements()*3];

		delete [] density;
		density = new T[domain_cells.elements()];

		delete [] flags;
		flags = new int[domain_cells.elements()];

		domain_scale[0] = 1.0/(float)domain_cells.max();
		domain_scale[1] = 1.0/(float)domain_cells.max();
		domain_scale[2] = 1.0/(float)domain_cells.max();
	}

	/**
	 * render the cube using vectors
	 * \param increment	incrementor specified how many vectors should be ommited in a dimension
	 */
	void render(int increment = -1)
	{
		// load array from gpu
		cLbmOpencl->storeVelocity(velocity);
		cLbmOpencl->storeDensity(density);
		cLbmOpencl->storeFlags(flags);

		// use incrementor to omit rendering points
		if (increment == -1)
			increment = domain_cells[0]/16;
		if (increment <= 0)
			increment = 1;

		int increment_half = increment / 2;

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();

			glTranslatef(	-(float)domain_cells[0]/(float)domain_cells.max()*0.5,
							-(float)domain_cells[1]/(float)domain_cells.max()*0.5,
							-(float)domain_cells[2]/(float)domain_cells.max()*0.5
					);
			glScalef(domain_scale[0], domain_scale[1], domain_scale[2]);
			glTranslatef(0.5f, 0.5f, 0.5f);	// move points to center of cell

			// use incrementor to omit rendering points
			size_t a = 0;
			T *velx = velocity;
			T *vely = velocity+domain_cells.elements();
			T *velz = velocity+domain_cells.elements()*2;

			// draw dots
			glBegin(GL_POINTS);
				glColor3f(1,1,1);
				for (int z = increment_half; z < domain_cells[2]; z += increment)
					for (int y = increment_half; y < domain_cells[1]; y += increment)
						for (int x = increment_half; x < domain_cells[0]; x += increment)
							glVertex3f((float)x, (float)y, (float)z);
			glEnd();
#if 0
			T max_length = 1;
			T vector_scale = 1;
#else
			T max_length = 0;
			for (int a = 0; a < domain_cells.elements(); a++)
			{
				if (flags[a] != LBM_FLAG_FLUID)
					continue;

				T val = velx[a]*velx[a] + vely[a]*vely[a] + velz[a]*velz[a];
				if (max_length < CMath<T>::abs(val))
					max_length = CMath<T>::abs(val);
			}

			max_length = CMath<T>::sqrt(max_length);
			T vector_scale;
			if (max_length < 0.00000001)
				max_length = 0.00000001;
#endif
			vector_scale = ((T)1.0/max_length);

			// draw velocity vectors
			glBegin(GL_LINES);
			for (int z = increment_half; z < domain_cells[2]; z += increment)
			{
				for (int y = increment_half; y < domain_cells[1]; y += increment)
				{
					for (int x = increment_half; x < domain_cells[0]; x += increment)
					{
						a = x + y*domain_cells[0] + z*domain_cells[0]*domain_cells[1];


						CVector<3,float> vel;
						vel[0] = velx[a];
						vel[1] = vely[a];
						vel[2] = velz[a];

						float length = vel.length()*vector_scale;

						if (length > 0.00000000001)
						{
							vel *= vector_scale;

							//glColor3f(1,1,z/(float)domain_cells[2]);
							glColor3f(0, 0, 1);
							glVertex3f((float)x, (float)y, (float)z);

							#if 1
								vel *= 0.5f/length;	// normalize
								vel *=  (T)increment;	// scale vectors according to visualization incrementation

								//glColor3f(1,0,z/(float)domain_cells[2]);
								glColor3f(1.0-((1.0-length)*(1.0-length)) , 0, 0);

								glVertex3f(	(float)x + vel[0],
										(float)y + vel[1],
										(float)z + vel[2]
										);
							#else
								vel.normalize();

								glColor3f(length/max_length,0,0);
								glVertex3f(	(float)x + vel[0],
										(float)y + vel[1],
										(float)z + vel[2]
										);
							#endif
						}
					}
				}
			}
			glEnd();

		glPopMatrix();
	}

	~CLbmVisualization()
	{
		if (velocity)
			delete velocity;

		if (density)
			delete density;

		if (flags)
			delete flags;
	}
};

#endif
