/*
 * Copyright 2010 Martin Schreiber
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#ifndef CEYEBALL_HPP
#define CEYEBALL_HPP

#include "libmath/CVector3.hpp"
#include "libmath/CMatrix.hpp"

/**
 * class to handle the movement on a 3d ball
 * the eye should rotate around the 3d ball and rotate around it's local x and y axis at the local position
 * the direction of the x and y axis in sphere coordinates is then given by the up / right coordinate
 *
 * the using code just has to rotate everything around the cartesian x, y and z axis according to the rotation degrees given by
 * vec3f rotate.
 * the rotation ordering is important: Z Y and then X
 *
 * TODO: this does not work as it should do but we can rotate somehow
 */
template <typename T>
class CEyeBall
{
private:
public:
	// all vectors have length 1
	vec3f eye;		///< eye look direction - always to center
	vec3f up;		///< up vector
	vec3f right;		///< vector to the right

	/*
	 * rotation angle given in radians
	 * 1st angle:	rotation around Y axis (alpha rotation)
	 * 2nd angle:	rotation around X axis (beta rotation)
	 * 3rd angle:	rotation around Z axis (gamma rotation)
	 *
	 * glRotate3f(rotateAngle[2], 0, 0, 1);
	 * glRotate3f(rotateAngle[1], 1, 0, 0);
	 * glRotate3f(rotateAngle[0], 0, 1, 0);
	 */
	/*
	vec3f rotationAngleDegrees;	///< the rotation components for x, y and z rotation in degree
	vec3f rotationAngleRadians;	///< the rotation components for x, y and z rotation in radians
	*/

	float rotationMatrix[4][4];

	CEyeBall()
	{
		reset();
	}

	void reset()
	{
		eye = vec3f(0, 0, 1);
		up = vec3f(0, 1, 0);
		right = vec3f(1, 0, 0);
/*
		rotationAngleDegrees = vec3f(0,0,0);
		rotationAngleRadians = vec3f(0,0,0);
*/
		reconstructRotationMatrix();
	}

	void makeNumericStable()
	{
		right = up.crossProd(eye);
		up = eye.crossProd(right);

		// TODO: use cross products to preserve 90 degree angles
		eye.normalize();
		eye.clamp1_1();

		up.normalize();
		up.clamp1_1();

		right.normalize();
		right.clamp1_1();
	}

	/**
	 * return the correct angle build by perpendicular lying vectors with length x and y
	 */
	T getATan(T x, T y, T extra_add = 0)
	{
		T v = y / x;
		T angle = atan(v);	// \in [-PI_2;PI_2]

		if (x < 0)
		{
			if (y < 0)	angle += -M_PI;
			else		angle += M_PI;
		}

		angle += extra_add;	// (0,0,1) is the origin eye position

		if (angle < -M_PI)	angle += 2*M_PI;
		else if (angle > M_PI)	angle -= 2*M_PI;

		return angle;
	}

	/**
	 * reconstruct angles given by the vectors eye and up
	 * the right vector is not necessary
	 *
	 * the reconstruction works by 3 steps:
	 * 1) computate angle alpha for rotation around the y axis until the eye point is in the middle vertical line
	 * 2) computate angle beta for rotation around new x axis ultil the eye point is in center
	 * 3) rerotate the up vector with inverted alpha and beta values + computate difference between original up and rerotated up vector for angle gamma
	 */
#if 0
	void reconstructAngles()
	{
		cerr << "not implemented yet!" << endl;
	// ALPHA
		/*
		 * rotation around y axis is given by x and z component of the up vector
		 * we cannot use the eye vector because we have to rotate to the position at the circuit where to rotate up
		 *
		 * if up aims down, add 180 degree
		 */
		rotationAngleRadians[0] = getATan(eye[0], eye[2], -M_PI_2 + (up[1] < 0 ? M_PI : 0));

		/*
		 * if the y component of the up vector is negative, we look upside down and have to
		 * add 180 degree to the x rotation!
		 */

	// BETA
		T d = sqrt(eye[0]*eye[0]+eye[2]*eye[2]);
		if (up[1] < 0)		// if the y component of the up vector is negative, we are on the opposite side
			d = -d;

		rotationAngleRadians[1] = getATan(d, eye[1]);

	// GAMMA
		// now we use the previous computated angles to rotate the up vector back to the origin
		vec3f back_up;
		matrix3f m;

		m.genRotation(-rotationAngleRadians[0], vec3f(0, 1, 0));
		back_up = m*up;

		m.genRotation(-rotationAngleRadians[1], vec3f(1, 0, 0));
		back_up = m*up;

//		cout << eye << "      " << up << "          " << back_up << endl;

		// use simplyfied dot product with vec3f(0, 1, 0);
		if (back_up[1] > 1)		rotationAngleRadians[2] = 0;
		else if (back_up[1] < -1)	rotationAngleRadians[2] = M_PI;
		else	rotationAngleRadians[2] = acos(back_up[1]);
	
		if (back_up[0] < 0)	rotationAngleRadians[2] = -rotationAngleRadians[2];

		// reconstruct angles in degrees
		rotationAngleDegrees[0] = rotationAngleRadians[0]*(180/M_PI);
		rotationAngleDegrees[1] = rotationAngleRadians[1]*(180/M_PI);
		rotationAngleDegrees[2] = rotationAngleRadians[2]*(180/M_PI);

//		cout << rotationAngleDegrees << "      " << eye << endl;
	}
#endif

	void reconstructRotationMatrix()
	{
		rotationMatrix[0][0] = right[0];
		rotationMatrix[1][0] = right[1];
		rotationMatrix[2][0] = right[2];
		rotationMatrix[3][0] = 0;

		rotationMatrix[0][1] = up[0];
		rotationMatrix[1][1] = up[1];
		rotationMatrix[2][1] = up[2];
		rotationMatrix[3][1] = 0;

		rotationMatrix[0][2] = eye[0];
		rotationMatrix[1][2] = eye[1];
		rotationMatrix[2][2] = eye[2];
		rotationMatrix[3][2] = 0;

		rotationMatrix[0][3] = 0;
		rotationMatrix[1][3] = 0;
		rotationMatrix[2][3] = 0; 
		rotationMatrix[3][3] = 1;
	}

	/**
	 * rotate around given angles - the angles are given in degrees
	 * the degree parameters specify the rotation of a viewed object which lies in the center
	 *
	 * alpha:	rotation around "screen" y axis
	 * 		if positive, rotate rotate object clockwise
	 *
	 * beta:	rotation around "screen" x axis
	 * 		if positive, rotate object up
	 *
	 * gamma:	rotation around "screen" z axis
	 * 		if positive, rotate anti-clockwise (clockwise rotation of object)
	 */
	void rotate(	T alpha,
			T beta,
			T gamma)
	{
		CMatrix3<float> m;

//		cout << eye << "      " << endl;
		// alpha: rotate eye and right vector around up vector
		m.genRotation(-alpha*(M_PI/180), up);
		eye = m*eye;
		right = m*right;

		// beta: rotate up around right vector
		m.genRotation(-beta*(M_PI/180), right);
		eye = m*eye;
		up = m*up;

		// gamma: rotate up around right vector
		m.genRotation(-gamma*(M_PI/180), eye);
		right = m*right;
		up = m*up;

		makeNumericStable();

//		reconstructAngles();
		reconstructRotationMatrix();
//		cout << eye << "      " << endl;
	}


	/**
	 * rotate vector with given alpha, beta and gamma parameters
	 */
	/*
	vec3f rotate(	T alpha,
			T beta,
			T gamma,
			vec3f vec
			)
	{
		matrix3f m;

		vec3f neye = eye;
		vec3f nup = up;
		vec3f nright = right;

		// alpha: rotate eye and right vector around up vector
		m.genRotation(-alpha*(M_PI/180), nup);
		neye = m*neye;
		nright = m*nright;

		// beta: rotate up around right vector
		m.genRotation(-beta*(M_PI/180), nright);
		neye = m*neye;
		nup = m*nup;

		// gamma: rotate up around right vector
		m.genRotation(-gamma*(M_PI/180), neye);
		nright = m*nright;
		nup = m*nup;

		return neye;
	}
	*/


	/**
	 * rotate around basis given by rball
	 */
	/*
	void rotate(	T alpha,
			T beta,
			T gamma,
			CEyeBall &rball
			)
	{
		matrix3f m;

		// alpha: rotate eye and right vector around up vector
		m.genRotation(-alpha*(M_PI/180), rball.up);
		eye = m*eye;
		up = m*up;
		right = m*right;

		// beta: rotate up around right vector
		m.genRotation(-beta*(M_PI/180), rball.right);
		eye = m*eye;
		up = m*up;
		right = m*right;

		// gamma: rotate up around right vector
		m.genRotation(-gamma*(M_PI/180), rball.eye);
		eye = m*eye;
		up = m*up;
		right = m*right;

		cout << rball.eye << "      " << rball.up << "         " << rball.right << "     " << eye << endl;
//		cout << eye << "      " << rball.right << "  " << beta << endl;

//		makeNumericStable();

//		reconstructAngles();
		reconstructRotationMatrix();
	}
	*/

};


#endif
